# Copyright (C) 2020 
#
# The source code in this file is released under the GPLv3.
# Go to https://www.gnu.org/licenses/gpl.txt for the full license details.
#
import upygame as pygame
import urandom as random
import umachine
import gc

import sfx
import res
import maps
from states import *


# Set palette
pygame.display.set_palette_16bit([
    0, 6438, 18917, 10825, 47398, 688, 41764, 17475,
    0xdc9c, 13598, 60486, 40179, 42596, 0xd6fb, 63245, 65535
]);


GO_START = 0
GO_LEVELS = 1
GO_INTRO = 2
GO_OUTRO = 3
GO_LOGO = 10
GO_CATSCENE = 42


def level_loop(screen):
    level_idx = maps.IDX_START
    sfx.play_bgm("vocfi/stora.wav")
    while level_idx < len(maps.levels):
        cur_level = maps.levels[level_idx]
        state = MapState(screen, cur_level.width, cur_level.height, cur_level.map_data)
        exit_status = state.mainloop()
        state = None  # repeat level
        gc.collect()

        if exit_status == 0:
            if level_idx == maps.IDX_LAST_STORAGE_LEVEL: # last storage level
                sfx.play_bgm("vocfi/sneak.wav")
                state = CatScene(screen) # show cutscene
                state.mainloop()
                sfx.play_bgm("vocfi/offic.wav") ## office levels
            level_idx += 1  # load next level
        elif exit_status == 1:
            break # goto start menu

    if level_idx == len(maps.levels):
        return 3
    return 0
    
# start game
def main():
    pygame.display.init()
    screen = pygame.display.set_mode()
    exit_status = GO_LOGO

    while True:
        if exit_status == GO_LOGO:
            state = LogoScreen(screen)
            state.mainloop()
            exit_status = GO_START
        elif exit_status == GO_OUTRO:
            state = OutroScreen(screen)
            state.mainloop()
            exit_status = GO_START
        elif exit_status == GO_LEVELS:
            exit_status = level_loop(screen)
        elif exit_status == GO_INTRO:
            state = IntroScreen(screen)
            state.mainloop()
            exit_status = GO_LEVELS
        elif exit_status == GO_CATSCENE:
            state = CatScene(screen)
            state.mainloop()
            exit_status = GO_START
        else:
            state = StartScreen(screen)
            state.mainloop()
            exit_status = GO_INTRO

main()