# Copyright (C) 2020 
#
# The source code in this file is released under the GPLv3.
# Go to https://www.gnu.org/licenses/gpl.txt for the full license details.
#
import upygame as pygame
import urandom as random
import umachine
import gc

import res
import maps
import sfx
from drawables import *
import dialog

class GameState:
    screen = None
    u_pressed = False
    d_pressed = False
    l_pressed = False
    r_pressed = False
    a_pressed = False
    b_pressed = False
    exit_status = None
    fading = False
    
    def __init__(self, screen):
        gc.collect()
        self.screen = screen

    def check_events(self):
        eventtype = pygame.event.poll()
        if self.fading:
            return
        
        if eventtype == pygame.NOEVENT:
            pass
        elif eventtype.type == pygame.KEYDOWN:
            if eventtype.key == pygame.K_RIGHT:
                self.r_pressed = True
            elif eventtype.key == pygame.K_LEFT:
                self.l_pressed = True
            elif eventtype.key == pygame.K_UP:
                self.u_pressed = True
            elif eventtype.key == pygame.K_DOWN:
                self.d_pressed = True
            elif eventtype.key == pygame.BUT_A:
                self.a_pressed = True
            elif eventtype.key == pygame.BUT_B:
                self.b_pressed = True

        elif eventtype.type == pygame.KEYUP:
            if eventtype.key == pygame.K_RIGHT:
                self.r_pressed = False
            elif eventtype.key == pygame.K_LEFT:
                self.l_pressed = False
            elif eventtype.key == pygame.K_UP:
                self.u_pressed = False
            elif eventtype.key == pygame.K_DOWN:
                self.d_pressed = False
            elif eventtype.key == pygame.BUT_A:
                self.a_pressed = False
            elif eventtype.key == pygame.BUT_B:
                self.b_pressed = False
            elif eventtype.key == pygame.BUT_C:
                sfx.disable_audio()

    def fade(self, dir=1):
        FADE_EFFECT.setup(dir)
        self.fading = True
        
    def update_logic(self):
        pass

    def update_screen(self):
        if self.fading:
            self.fading = FADE_EFFECT.update()
            FADE_EFFECT.draw(self.screen)
        pygame.display.flip()
    
    def mainloop(self):
        running = True
        while running:
            self.check_events()
            self.update_logic()
            self.update_screen()
            if self.exit_status is not None:
                break
        return self.exit_status


zk_logo = pygame.surface.Surface(110, 88, res.pixels_zk)
class LogoScreen(GameState):
    step = 0
    
    def __init__(self, screen):
        GameState.__init__(self, screen)
    
    def update_screen(self):
        self.screen.blit(zk_logo, 0, 0)
        if self.step == 0:
            self.step += 1
            self.fade(-3)
        elif self.step == 1 and not self.fading:
            sfx.play_sfx(sfx.meow)
            umachine.wait(1000)
            self.fade(3)
            self.step += 1
        elif self.step == 2 and not self.fading:
            self.screen.fill(0, pygame.Rect(0, 0, 110, 88)) 
            self.exit_status = 0 # GO_START
        GameState.update_screen(self)
        

intro_a_button = pygame.surface.Surface(8, 8, res.pixels_a_button)
top_bg = pygame.surface.Surface(110, 88, res.top_bgPixels)
top_railing = pygame.surface.Surface(110, 5, res.top_railingPixels)
top_balloon = pygame.surface.Surface(50, 72, res.top_balloonPixels)
class StartScreen(GameState):
    blinker = 0
    
    def __init__(self, screen):
        GameState.__init__(self, screen)
        sfx.stop()
    
    def update_screen(self):
        self.screen.blit(top_bg, 1, 1)
        self.screen.blit(top_balloon, 1, 9, 12)
        self.screen.blit(top_railing, 1, 76)
        
        umachine.draw_text(61, 20, "The Villainy", 15)
        umachine.draw_text(61, 26, "      of    ", 15)
        umachine.draw_text(61, 32, "Cat Food Inc.", 15)
        
        self.blinker += 1
        if self.blinker > 40:
            umachine.draw_text(37, 60, "Press", 15)
            self.screen.blit(intro_a_button, 60, 58)
        if self.blinker > 80:
            self.blinker = 0
        GameState.update_screen(self)
    
    def check_events(self):
        GameState.check_events(self)
        if self.a_pressed or self.b_pressed:
            self.exit_status = 1

intro_guy = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_guy_stand_big)
intro_guy_cat = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_guy_go_cat)
intro_guy_up = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_guy_go_up)
intro_guy_ohno = pygame.surface.Surface(res.big_guy_w, res.big_guy_h + 4, res.pixels_guy_stand_big_oh_no)
intro_johnny = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_johnny_stand_big)
intro_evilguy = pygame.surface.Surface(res.evilguy_w, res.evilguy_h, res.pixels_evilguy_stand_big)
intro_evilguy_catless = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_evilguy_stand_nocat)
intro_nanji = pygame.surface.Surface(8, 8, res.pixels_nanji)
intro_nanji2 = pygame.surface.Surface(8, 8, res.pixels_nanji2)
intro_house = pygame.surface.Surface(64, 32, res.housePixels)
intro_catfoodinc = pygame.surface.Surface(90, 37, res.catfoodincPixels)
intro_guy_lookup = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_guy_stand_big_lookup)
intro_lobby = pygame.surface.Surface(90, 37, res.lobby_finalPixels)
class IntroScreen(GameState):
    step = -1
    final_step = 1
    anim_timer = 0
    x_offset = 0
    in_dialog = False
    dialog = dialog.Dialog(1, 60, 110, 33, dialog.INTRO_DLG)
    walkpos = 0  

    def update_screen(self):
        if self.exit_status is not None:
            self.screen.fill(0, pygame.Rect(0, 0, 110, 88)) 
            GameState.update_screen(self)
            return
        
        if self.step < 10: 
            self.screen.fill(9, pygame.Rect(10, 15, 90, 35)) 
            self.screen.fill(7, pygame.Rect(10, 50, 90, 2)) 
            self.screen.blit(intro_house, -32 + self.x_offset, 19)
            self.screen.blit(intro_guy, -40 + self.x_offset, 38)
            self.screen.blit(intro_johnny, -50 + self.x_offset, 38, 12, self.step > 1) # green=transparent
        
        # First of all: Fade in
        if self.step == -1:
            self.step = 0
            sfx.play_bgm("vocfi/intro.wav")
            self.screen.blit(intro_nanji, 80 + self.x_offset, 42)
            self.fade(-2)
        # Step 0: Show cat. Pan to the left. Show Guy and Johnny. 
        elif self.step == 0:
            if self.x_offset > 0 or self.anim_timer == 40:
                self.x_offset += 1
            if self.x_offset > 60:
                self.step += 1
                self.in_dialog = True

            if self.anim_timer > 20:
                self.screen.blit(intro_nanji, 80 + self.x_offset, 42)
            else:
                self.screen.blit(intro_nanji2, 80 + self.x_offset, 42)
        # Step 1: Show wonderful dialog
        elif self.step == 1 and not self.in_dialog:
            self.dialog.increase()
            self.step += 1
        # Step 2: Pan back to the right - oh, no! Catnappers!
        elif self.step == 2:
            self.x_offset -= 1
            self.screen.blit(intro_evilguy, 75 + self.x_offset, 38, 12) # green=transparent
            if self.x_offset <= 0:
                sfx.play_sfx(sfx.meow)
                self.step += 1
                self.walkpos = 75
                sfx.play_bgm("vocfi/drama.wav")
        # Step 3: Evilguy walks off with catnapped kitteh
        elif self.step == 3:
            y_offset = self.walkpos % 2
            self.screen.blit(intro_evilguy, self.walkpos + self.x_offset, 37 + y_offset, 12) # green=transparent
            # really?! how about making an animated sprite class that keeps
            # track of positioning and anim timer instead of this mess? O___o
            self.walkpos += 1
            if self.walkpos > 110:
                self.step += 1
                self.walkpos = 0
        # Step 4: Guy enters from left
        elif self.step == 4:
            y_offset = self.walkpos % 2
            self.screen.blit(intro_guy, -20 + self.walkpos + self.x_offset, 37 + y_offset, 0, True)
            self.walkpos += 1
            if self.walkpos > 50:
                self.step += 1
                self.in_dialog = True
        # Step 5: Oh no!!
        elif self.step == 5 and self.in_dialog:
            self.screen.blit(intro_guy_ohno, -20 + self.walkpos + self.x_offset, 34)
        # Step 5+ (after dialog): Fadeout
        elif self.step == 5 and not self.in_dialog:
            self.dialog.increase()
            self.screen.blit(intro_guy_ohno, -20 + self.walkpos + self.x_offset, 34)
            self.fade(2)
            self.step += 1
        elif self.step == 6 and self.fading:
            self.screen.blit(intro_guy_ohno, -20 + self.walkpos + self.x_offset, 34)
        elif self.step == 6 and not self.fading:
            self.screen.fill(0, pygame.Rect(10, 15, 90, 38))
            self.fade(-2)
            self.step = 10
        # Fadein Cat Food inc. building 
        elif self.step == 10:
            self.screen.blit(intro_catfoodinc, 10, 16)
            self.screen.blit(intro_guy, 10, 38, 0, True)
            if not self.fading:
                self.step += 1
                self.in_dialog = True
        elif self.step == 11:  # I'll rescue you, Nanji!
            self.screen.blit(intro_catfoodinc, 10, 16)
            self.screen.blit(intro_guy_lookup, 10, 38, 0, True)
            if not self.in_dialog:
                self.fade(2)
                self.step += 1
        elif self.step == 12:
            if self.fading:
                self.screen.blit(intro_catfoodinc, 10, 16)
                self.screen.blit(intro_guy_lookup, 10, 38, 0, True)
            else:
                self.step += 1
                self.walkpos = 10
                self.fade(-2)
                sfx.play_bgm("vocfi/sneak.wav")
        # lobby
        elif self.step == 13:
            self.screen.blit(intro_lobby, 10, 16)
            if self.fading:
                self.screen.blit(intro_guy, 10 + self.walkpos, 29, 0, True)
            else:
                y_offset = self.walkpos % 2
                if self.walkpos > 70:
                    y_offset += 3
                elif self.walkpos > 60:
                    y_offset += 1
                self.walkpos += 1
                self.screen.blit(intro_guy, 10 + self.walkpos, 29 + y_offset, 0, True)
                if self.walkpos > 118:
                    self.step += 1
                    self.fade(2)
        elif self.step == 14:
            if self.fading:
                self.screen.blit(intro_lobby, 10, 16)
            else:
                self.step += 1
                self.exit_status = 1
        
        # dialog
        if self.in_dialog:
            self.dialog.show_line()
            self.screen.blit(intro_a_button, 102, 80)

        #draw box
        self.screen.fill(0, pygame.Rect(0, 15, 10, 40))
        self.screen.fill(0, pygame.Rect(100, 15, 10, 40))
        
        self.anim_timer += 1        
        if self.anim_timer > 40:
            self.anim_timer = 0

        GameState.update_screen(self)
    
    def check_events(self):
        GameState.check_events(self)
        if self.a_pressed and self.in_dialog:
            self.a_pressed = False
            self.in_dialog = self.dialog.increase()
        elif self.b_pressed:
            self.exit_status = 1


class CatScene(GameState):
    step = -1
    walkpos = 90
    turn_johnny = True
    turn_evilguy = True
    guy_image = intro_guy_cat
    
    def update_screen(self):
        umachine.wait(25)
        
        self.screen.blit(intro_lobby, 10, 16)
        self.screen.blit(intro_evilguy_catless, 15, 30, 12, self.turn_evilguy)
        self.screen.blit(intro_johnny, 27, 30, 12, self.turn_johnny)
        if self.guy_image:
            self.screen.blit(self.guy_image, self.walkpos, 32 - self.walkpos % 2)
            
        if self.step == -1:
            self.step = 0
            self.fade(-2)
        elif self.step == 0 and not self.fading:
            self.step += 1
        elif self.step == 1:
            self.walkpos -= 1
            if self.walkpos <= 60:
                self.step += 1
        elif self.step == 2:
            self.walkpos += 1
            if self.walkpos >= 65:
                self.guy_image = intro_guy_up
                self.step += 1
        elif self.step <= 5 and self.step > 2:
            self.step += 1
        elif self.step == 6:
            self.guy_image = None
            self.turn_johnny = False
            self.step += 1
        elif self.step <= 10 and self.step > 6:
            self.step += 1
        elif self.step == 11:
            self.turn_evilguy = False
            self.step += 1
        elif self.step <= 14 and self.step > 11:
            self.step += 1
        elif self.step == 15:
            self.fade(2)
            self.step += 1
        elif self.step == 16 and not self.fading:
            self.screen.fill(0, pygame.Rect(1, 1, 110, 88))
            self.exit_status = 1
        GameState.update_screen(self)
    
    def check_events(self):
        GameState.check_events(self)
        if self.b_pressed:
            self.exit_status = 1

outro_guy = pygame.surface.Surface(res.big_guy_w, res.big_guy_h, res.pixels_guy_stand_cat)
top_door_open = pygame.surface.Surface(22, 18, res.top_door_openPixels)
cloud = pygame.surface.Surface(52, 23, res.pixels_cloud)
cloud2 = pygame.surface.Surface(20, 10, res.pixels_cloud2)
class OutroScreen(GameState):
    step = 0
    draw_door = False
    guy_pos_x = 84
    guy_pos_y = 57
    evilguy_pos_x = 84
    johnny_pos_x = 84
    balloon_pos_x = 1
    balloon_pos_y = 9
    cloud_pos_x = 10
    cloud_pos_y = -44
    cloud2_pos_x = 50
    cloud2_pos_y = -80
    top_pos_y = 1
    
    def update_screen(self):
        #umachine.wait(25)
        self.screen.fill(9, pygame.Rect(1, 1, 110, 88))
        self.screen.blit(top_bg, 1, self.top_pos_y)
        self.screen.blit(cloud, self.cloud_pos_x, self.cloud_pos_y)
        self.screen.blit(cloud2, self.cloud2_pos_x, self.cloud2_pos_y)
        if self.draw_door:
            self.screen.blit(top_door_open, 78, 58)
        
        if self.step == 0: # fadein
            sfx.play_bgm("vocfi/drama.wav")
            self.step += 1
            self.fade(-2)
        elif self.step == 1 and not self.fading:
            self.step += 1
        elif self.step == 2:  # door opens
            umachine.wait(1000)
            self.draw_door = True
            self.step += 1
        elif self.step == 3: # guy appears, moves left
            self.screen.blit(intro_guy_cat, self.guy_pos_x,64 + self.guy_pos_x % 2)
            self.guy_pos_x -= 1
            if self.guy_pos_x < 50:  # evilguy appears, moves left
                self.screen.blit(intro_evilguy_catless, self.evilguy_pos_x, 64 + self.evilguy_pos_x % 2, 12, True)
                self.evilguy_pos_x -= 1
            if self.guy_pos_x < 37:
                self.step += 1
        elif self.step == 4: # johnny appears. the suspense is killing me..
            self.screen.blit(intro_guy_cat, self.guy_pos_x, 64 + self.guy_pos_x % 2, 0)
            self.screen.blit(intro_evilguy_catless, self.evilguy_pos_x, 64 + self.evilguy_pos_x % 2, 12, True)
            self.screen.blit(intro_johnny, self.johnny_pos_x, 64 + self.johnny_pos_x % 2, 12, True)
            self.johnny_pos_x -= 1
            self.evilguy_pos_x -= 1
            self.guy_pos_x -= 1
            if self.guy_pos_x <= 21:
                self.step += 1
        elif self.step == 5:  # balloon flies up
            self.screen.blit(outro_guy, self.guy_pos_x, self.guy_pos_y, 0, True)
            self.screen.blit(intro_evilguy_catless, self.evilguy_pos_x, 64 + self.evilguy_pos_x % 2, 12, True)
            self.screen.blit(intro_johnny, self.johnny_pos_x, 64 + self.johnny_pos_x % 2, 12, True)
            
            if self.evilguy_pos_x > 25:
                self.johnny_pos_x -= 1
                self.evilguy_pos_x -= 1
            
            self.guy_pos_y -= 1 
            self.balloon_pos_y -= 1
            if self.guy_pos_y < 50:
                self.balloon_pos_x -= 1
                self.guy_pos_x -= 1
            if self.balloon_pos_x <= -50:
                sfx.stop()
                self.step += 1
                
        elif self.step == 6: # johnny & evilguy go home
            if self.johnny_pos_x < 84:
                self.johnny_pos_x += 1
                self.screen.blit(intro_johnny, self.johnny_pos_x, 64 + self.johnny_pos_x % 2, 12, False)
            if self.evilguy_pos_x < 84:
                self.evilguy_pos_x += 1
                self.screen.blit(intro_evilguy_catless, self.evilguy_pos_x, 64 + self.evilguy_pos_x % 2, 12, False)
            else:
                self.draw_door = False
                sfx.play_sfx(sfx.push)
                self.step += 1
        elif self.step >= 7:
            self.screen.blit(outro_guy, self.guy_pos_x, self.guy_pos_y, 0, True)
        self.screen.blit(top_balloon, self.balloon_pos_x, self.balloon_pos_y, 12)
        self.screen.blit(top_railing, 1, self.top_pos_y + 76)

        if self.step == 7:
            self.top_pos_y += 1
            self.cloud_pos_y += 1
            self.cloud2_pos_y += 1
            if self.top_pos_y > 88:
                sfx.play_bgm("vocfi/outro.wav")
                self.step += 1
                self.balloon_pos_y = -80
                self.balloon_pos_x = 30
                self.guy_pos_x = 50
                self.guy_pos_y = -30
        elif self.step == 8:
            self.cloud_pos_y += 1
            self.cloud2_pos_y += 1
            if self.balloon_pos_y < 10:
                self.balloon_pos_y += 1
                self.guy_pos_y += 1
            else:
                umachine.draw_text(40, 20, "The End", 15)
            if self.cloud_pos_y > 150:
                self.cloud_pos_y = -24
                self.cloud_pos_x = 10 + random.getrandbits(6)
            if self.cloud2_pos_y > 140:
                self.cloud2_pos_y = -30
                self.cloud2_pos_x = 30 + random.getrandbits(6)
        elif self.step == 9:
            if not self.fading:
                self.screen.fill(0, pygame.Rect(1, 1, 110, 88))
                self.exit_status = 0

        
        GameState.update_screen(self)
    
    def check_events(self):
        GameState.check_events(self)
        if self.a_pressed or self.b_pressed:
            if self.step == 8:
                self.step += 1
                self.fade(2)


game_b_button = pygame.surface.Surface(8, 8, res.pixels_b_button)
class MapState(GameState):

    def __init__(self, screen, map_w, map_h, map_data):
        self.cur_map = Map(map_w, map_h, map_data)
        self.exit_coord = None
        self.player = Player()
        self.sprites = []
        self.crates = []
        self.doors = []
        self.keys = []
        self.cats = []
        self.button = None
        self.btn_door = None
        self.trigger = None
        self.enemies = ()
        self.start_coord = None

        # find player, crates, doors, buttons
        idx = 0
        for tile in self.cur_map.map_data:
            row = idx // self.cur_map.width
            column = idx % self.cur_map.width
            if tile in ("e", "E"):
                self.exit_coord = (column, row)
            elif tile in ("c", "C"):
                self.crates.append(Crate(row, column, tile))
            elif tile == "s":
                self.player.set_coord(row + 1, column)
                self.start_coord = (column, row)
            elif tile == "S":
                self.player.set_coord(row - 1, column)
                self.player.cur_image_idx = 1
                self.start_coord = (column, row)
            elif tile == "o":
                self.btn_door = ButtonDoor(row, column)
                self.sprites.append(self.btn_door)
            elif tile == "b":
                self.button = Button(row, column)
                self.sprites.append(self.button)
            elif tile == "d":
                self.doors.append(Door(row, column))
            elif tile == "k":
                self.keys.append(Key(row, column))
            elif tile == "n":
                self.cats.append(Cat(row, column))
            elif tile == 't':
                self.trigger = (column, row)
            idx += 1

        self.sprites.extend(self.doors)
        self.sprites.extend(self.keys)
        self.sprites.extend(self.crates)
        self.sprites.extend(self.cats)
        self.sprites.append(self.player) 
        sfx.play_sfx(sfx.step1)
        sfx.play_sfx(sfx.step2)
        sfx.play_sfx(sfx.step1)
        sfx.play_sfx(sfx.step2)
        GameState.__init__(self, screen)

    def check_events(self):
        GameState.check_events(self)
        if not self.player.keep_moving:
            if self.u_pressed:
                self.player.move(vy=-1)
            elif self.d_pressed:
                self.player.move(vy=1)
            elif self.l_pressed:
                self.player.move(vx=-1)
            elif self.r_pressed:
                self.player.move(vx=1)
            else: 
                self.player.stop()
        if self.b_pressed:
            self.exit_status = 2

    def update_screen(self):
        self.cur_map.draw(self.screen)
        for s in self.sprites:
            if s not in self.enemies:
                self.screen.blit(s.cur_image, s.x + self.cur_map.offset_x, s.y + self.cur_map.offset_y)
            else:
                self.screen.blit(s.cur_image, s.x + self.cur_map.offset_x, s.y + self.cur_map.offset_y, 8)
        
        # draw number of keys
        self.screen.blit(tiles["k"], 0, 0)
        umachine.draw_text(8, 3, "x" + str(self.player.keys), 15)
        
        # b to restart..
        umachine.draw_text(73, 82, "reset=", 15)
        self.screen.blit(game_b_button, 101, 80)
        GameState.update_screen(self)

    def contains_crate(self, column, row):
        for c in self.crates:
            if c.column == column and row == c.row:
                return c
        return None

    def contains_door(self, column, row):
        for d in self.doors:
            if d.column == column and d.row == row:
                return d
        return None
    
    def contains_key(self, column, row):
        for k in self.keys:
            if k.column == column and k.row == row:
                return k
        return None
    
    def contains_cat(self, column, row):
        for c in self.cats:
            if c.column == column and c.row == row:
                return c
        return None
    
    def contains_enemy(self, column, row):
        for e in self.enemies:
            if e.column == column and e.row == row:
                return e
        return None
    
    def win(self):
        self.exit_status = 0

    def update_ai(self, ai):
        check_vels = ai.get_turn_checks()
        for vel in check_vels:
            if not self.is_tile_blocked(ai.column + vel[0], ai.row + vel[1], vel[0], vel[1]):
                ai.move(vx=vel[0], vy=vel[1])
                break

    def is_tile_blocked(self, column, row, vx=0, vy=0):
        coord = (column, row)
        
        door = self.contains_door(column, row)
        if door and not door.activated:
            return True
        
        if self.enemies and (coord == (self.enemies[0].column, self.enemies[0].row) or coord == (self.enemies[1].column, self.enemies[1].row)):
            return True
        
        crate = self.contains_crate(column, row)
        if crate:
            next_coord = (column + vx, row + vy)
            return self.contains_crate(next_coord[0], next_coord[1]) or \
                self.cur_map.is_blocking(next_coord[0], next_coord[1]) or \
                (self.button.column, self.button.row) == (next_coord[0], next_coord[1]) or \
                coord == (next_coord[0], next_coord[1]) or \
                self.contains_door(next_coord[0], next_coord[1]) or \
                next_coord == (self.player.column, self.player.row) or \
                (self.enemies and (coord == (self.enemies[0].column, self.enemies[0].row) or coord == (self.enemies[1].column, self.enemies[1].row)))
            
        if self.cur_map.is_blocking(column, row) or coord == (self.button.column, self.button.row) or (coord == (self.btn_door.column, self.btn_door.row) and not self.btn_door.activated):
            return True
        return False

    def update_logic(self):
        # monstrosity... 
        
        for s in self.sprites:
            if s in self.enemies and self.player.reached_tile:
                self.update_ai(s)

            if s.vx or s.vy:
                new_row = s.row + s.vy
                new_col = s.column + s.vx
                stop_moving = False

                # check "collisions"
                if self.cur_map.is_blocking(new_col, new_row):
                    stop_moving = True
                elif self.button.row == new_row and self.button.column == new_col:
                    stop_moving = True
                    if not self.button.activated:
                        self.button.activate()
                        sfx.play_sfx(sfx.button)
                        self.btn_door.activate()
                elif not self.btn_door.activated and new_row == self.btn_door.row and new_col == self.btn_door.column:
                    stop_moving = True
                elif self.trigger and new_row == self.trigger[1] and new_col == self.trigger[0]:
                    self.trigger = None
                    johnny = Ai(self.start_coord[1] - 1, self.start_coord[0], True)
                    evilguy = Ai(self.start_coord[1] - 2, self.start_coord[0], False)
                    self.enemies = (johnny, evilguy)
                    self.sprites.append(johnny)
                    self.sprites.append(evilguy)

                # what a mess.. again...
                if not stop_moving:
                    door = self.contains_door(new_col, new_row)
                    if door and not door.activated:
                        stop_moving = True
                        if (s is self.player and self.player.keys > 0) or \
                            (s in self.enemies and s.keys > 0):
                            door.activate()
                            sfx.play_sfx(sfx.push)
                            s.keys -= 1
                    if not door:
                        crate = self.contains_crate(new_col, new_row)
                        if crate:
                            if s is self.player or s in self.enemies:
                                # check if crate moveable, push crate
                                next_next_row = new_row + s.vy
                                next_next_col = new_col + s.vx
                                if self.cur_map.is_blocking(next_next_col, next_next_row) or \
                                    self.contains_door(next_next_col, next_next_row) or \
                                    self.contains_crate(next_next_col, next_next_row) or \
                                    (next_next_row == self.btn_door.row and next_next_col == self.btn_door.column) or \
                                    (next_next_row == self.btn_door.row and next_next_col == self.btn_door.column) or \
                                    self.contains_enemy(next_next_col, next_next_row):
                                    stop_moving = True
                                else:
                                    crate.move(s.vx, s.vy)
                                    s.pushing_crate = crate
                                    
                            else:
                                # crates cannot push other crates!
                                stop_moving = True 
                                s.stop()

                if stop_moving:
                    s.set_coord(s.row, s.column)
                    s.vx = 0
                    s.vy = 0
                    s.stop()

                # moving objects destroy keys .. only player & enemies pick them up
                coll_key = self.contains_key(new_col, new_row)
                coll_cat = not coll_key and self.contains_cat(new_col, new_row)

                if coll_key:
                    self.sprites.remove(coll_key)
                    self.keys.remove(coll_key) # this is why spritegroups are so nice, you know...
                    if s is self.player or s in self.enemies:
                        s.keys += 1
                        sfx.play_sfx(sfx.pickup)
                elif coll_cat:
                    if s is self.player:
                        self.sprites.remove(coll_cat)
                        self.cats.remove(coll_cat) # again: this is why spritegroups are so nice, you know...
                        self.player.cats += 1
                        sfx.play_sfx(sfx.meow)
            
            # oh, yes. update the Sprite, pls
            s.update()

        if self.exit_coord == (self.player.column, self.player.row):
            self.win()
        elif self.contains_enemy(self.player.column, self.player.row):
            self.exit_status = 2
